package tp.impl;

import java.io.Serializable;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import tp.Customer;
import tp.InvalidTransactionException;
import tp.RID;
import tp.LM;
import tp.RID.Type;
import tp.RM;
import tp.TM;
import tp.Transaction;
import tp.TransactionAbortedException;

/**
 * Your Resource Manager.
 * 
 * Rudimentary implementation is provided for your convenience.
 * 
 * <h3>Command line arguments</h3>
 * 
 * The following arguments are further processed by this class
 * in addition to common arguments processed by {@link CSEP545ServiceObject#main(String[])}.
 * 
 * <ul>
 * <li><code>-x</code> or <code>--selfdestruct</code> <code>numio</code>:
 * Set the self destruction timer on this RM. If this argument is given,
 * the RM MUST call System.exit() <i>immediately</i> after numio disk operations.
 * (default: 0. no self destruction)</li>
 * <li><code>-t</code> or <code>--timeout</code> <code>s</code>: 
 * Set deadlock detection timeout to <code>s</code> seconds. 
 * (default: {@link tp.LM#DEFAULT_DEADLOCK_TIMEOUT})</li>
 * </ul>
 * 
 * @see tp.impl.CSEP545ServiceObject
 */
public class MyRM extends CSEP545ServiceObject implements RM {
    private static final long serialVersionUID = -824882060186317662L;

    /**
     * keep track of inventory information.
     */
    class Resource implements Serializable {
        private static final long serialVersionUID = 15980438631067294L;
        
        private RID rID;
        private int count;
        private int price;
        
        public Resource() {}
        public Resource(RID rID) {
            this.rID = rID;
        }
        public Resource(RID rID,int c,int p) {
            this.rID = rID;
            this.count = c;
            this.price = p;
        }
        
        public String getName() { return rID.getName(); }
        public RID getID() { return rID; }
        public RID.Type getType() { return rID.getType(); }
        public int getCount() { return count; }
        public int getPrice() { return price; }

        public void incrCount() { ++this.count; }
        public void incrCount(int c) { this.count += c; }
        
        public void decrCount(int c) { this.count -= c; }
        public void decrCount() { --this.count; }
        
        public void setCount(int count) { this.count = count; }
        public void setPrice(int price) { this.price = price; }
        
        @Override
        public int hashCode() { return rID.hashCode(); }

        @Override
        public String toString() {
            return rID.getName() + "," + count + "," + price;
        }
    }
    
    private transient volatile TM myTM;
    private transient LM myLM = new MyLM();
    
    /**
     * Inventory information. Should be part of your database.
     * Intentionally using {@link java.util.HashMap} to increase the chance of
     * incorrect synchronization using Lock Manager.
     */
    private HashMap<RID,Resource> resources = new HashMap<RID,Resource>();
    
    /**
     * Reservation information. Should be part of your database.
     * Intentionally using {@link java.util.HashMap} to increase the chance of
     * incorrect synchronization using Lock Manager.
     */
    private HashMap<Customer,Set<RID>> reservations = new HashMap<Customer,Set<RID>>();

    public MyRM() {
    }

    @Override
    public void commit(Transaction xid)
    throws InvalidTransactionException,
            TransactionAbortedException, RemoteException {
    }

    @Override
    public void abort(Transaction xid) throws RemoteException {
    }

    @Override
    public boolean add(Transaction xid, RID i, int count, int price)
            throws TransactionAbortedException, InvalidTransactionException,
            RemoteException {
        Resource ii = resources.get(i);
        if ( ii == null ) {
            resources.put(i, new Resource(i,count,price));
        } else {
            ii.incrCount(count);
            ii.setPrice(price);
            
            // TODO add locking code here
        }
        return true;
    }

    @Override
    public boolean delete(Transaction xid, RID rid)
            throws TransactionAbortedException, InvalidTransactionException,
            RemoteException {
        // TODO add locking code here
        Resource removed = resources.remove(rid);

        // drop all reservations on removed resource
        if ( removed != null ) {
            for ( Map.Entry<Customer,Set<RID>> e : reservations.entrySet() ) {
                e.getValue().remove(rid);
            }
        }
        
        return removed != null;
    }
    
    @Override
    public boolean delete(Transaction xid, RID i, int count)
            throws TransactionAbortedException, InvalidTransactionException,
            RemoteException {
        // TODO add locking code here
        Resource removed = resources.get(i);
        if ( removed == null ) {
            // silently discard
        } else {
            if ( removed.getCount() > count ) {
                removed.decrCount(count);
            } else {
                removed.setCount(0);
            }
        }
        return true;
    }

    @Override
    public int query(Transaction xid, RID rid)
            throws TransactionAbortedException, InvalidTransactionException,
            RemoteException {
        Resource resource = resources.get(rid);
        if ( resource == null ) {
            throw new IllegalArgumentException(rid+" does not exist");
        }
        return resource.getCount();
    }

    @Override
    public int queryPrice(Transaction xid, RID i)
    throws TransactionAbortedException, InvalidTransactionException, RemoteException {
        Resource ii = resources.get(i);
        if ( ii == null ) {
            throw new IllegalArgumentException(i+" does not exist");
        }
        return ii.getPrice();
    }

    @Override
    public String queryReserved(Transaction context, Customer customer)
            throws TransactionAbortedException, InvalidTransactionException,
            RemoteException {
        StringBuilder buf = new StringBuilder(512);
        
        Set<RID> reserved = reservations.get(customer);
        if ( reserved != null ) {
            for ( RID rid : reserved ) {
                if ( buf.length() > 0 ) {
                    buf.append(',');
                }
                buf.append(rid);
            }
        }
        return buf.toString();
    }

    @Override
    public int queryReservedPrice(Transaction context, Customer customer)
            throws TransactionAbortedException, InvalidTransactionException,
            RemoteException {
        int bill = 0;
        
        Set<RID> reserved = reservations.get(customer);
        if ( reserved != null ) {
            for ( RID rid : reserved ) {
                Resource r = resources.get(rid);
                if ( r == null ) {
                    throw new IllegalStateException(rid+" does not exist in RM");
                }
                bill += r.getPrice();
            }
        }
        
        return bill;
    }

    @Override
    public boolean reserve(Transaction context, Customer c, RID i)
            throws TransactionAbortedException, InvalidTransactionException,
            RemoteException {
        Resource ii = resources.get(i);
        
        if ( ii == null ) {
            throw new IllegalArgumentException(i+" does not exist!");
        }
        if ( ii.getCount() == 0 ) {
            return false;
        }
        
        Set<RID> r = reservations.get(c);
        if ( r == null ) {
            r = new LinkedHashSet<RID>();
            r.add(ii.getID());
            reservations.put(c, r);
        } else {
            r.add(ii.getID());
        }
        
        ii.decrCount();
        
        return true;
    }

    @Override
    public void unreserve(Transaction context, Customer c)
            throws TransactionAbortedException, InvalidTransactionException,
            RemoteException {
        Set<RID> r = reservations.get(c);
        if ( r == null ) {
            // silently discard
        } else {
            // TODO need to add lock
            for ( RID rid : r ) {
                Resource ii = resources.get(rid);
                if ( ii == null ) {
                    // FIXME warn that the rID does not exist!
                } else {
                    ii.incrCount();
                }
            }
            reservations.remove(c);
        }
    }

    @Override
    public String[] listResources(Transaction context, RID.Type type)
            throws TransactionAbortedException, InvalidTransactionException,
            RemoteException {
        ArrayList<String> result = new ArrayList<String>(resources.size());
        for ( Resource resource : resources.values() ) {
            if ( type == resource.getType() ) {
                result.add(resource.toString());
            }
        }
        return result.toArray(new String[0]);
    }

    @Override
    public Customer[] listCustomers(Transaction context)
    throws TransactionAbortedException, InvalidTransactionException,
            RemoteException {
        return reservations.keySet().toArray(new Customer[0]);
    }
    
    @Override
    public void shutdown() throws RemoteException {
        // TODO Do proper shutdown here
    }

    @Override
    public void selfDestruct(int diskWritesToWait) throws RemoteException {
        // TODO Set counter
    }
    
    @Override
    public String getName() throws RemoteException {
        return getServiceID().toString();
    }

    /**
     * @todo setup {@link #selfDestruct(int)} here
     */
    @Override
    protected void init(String[] args) throws Exception {
        int argc = 0;
        
        for ( ; argc < args.length; ++argc ) {
            String arg = args[argc];
            if ( "-x".equals(arg) || "--selfdestruct".equals(arg) ) {
                int nio = Integer.parseInt(args[++argc]);
                System.out.println("Setting self-destruction counters to "+nio);
                selfDestruct(nio);
            } else if ( "-t".equals(arg) || "--timeout".equals(arg) ) {
                int to = Integer.parseInt(args[++argc]);
                System.out.println("Setting deadlock timeout to "+to+" seconds");
                myLM.setDeadlockTimeout(to);
            } else {
                throw new IllegalArgumentException("unknown argument: "+arg);
            }
        }

        refresh();
    }

    @Override
    protected void initStorage() throws Exception {
        // TODO create database files, transaction logs
    }

    @Override
    protected void recovery() throws Exception {
        // TODO recover state from database file
    }

    @Override
    protected void startUp() throws Exception {
        // TODO deadlock detector, retry timeout
    }
    
    @Override
    protected void readyToServe() throws Exception {
        // register this resource manager to TM
        myTM.register(getName());
    }

    @Override
    public void refresh() throws RemoteException, NotBoundException {
        myTM = this.lookupRemote(ServiceID.forTM());
    }
}
